__doc__ = "Zope User Bulletin Board"

import Globals, OFS, Products, AccessControl, os

import ZUBB_files, ZUBB_methods

from ImageFile import ImageFile
from AccessControl import ClassSecurityInfo
from Products.ZCatalog import ZCatalog
from Acquisition import Implicit, aq_parent, aq_base 
from Globals import HTMLFile, Persistent
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from DateTime import DateTime
from Products.PythonScripts.standard import html_quote

emoticons_list = {'happy.gif':    (':-)', ':)', '=)'),
                  'unhappy.gif':  (':-(', ':(', '=(', ":'("),
                  'blah.gif':     (':-|', ':|', '=|', ':-I', ':I', '=|'),
                  'funny.gif':    (':-D', ':D', '=D'),
                  'confused.gif': (':-S', ':S', '=S', '=s', ':-s', ':s', ':-?'),
                  'ooh.gif':      (':-O', ':O', '=O', '=o', ':-o', ':o'),
                  'wink.gif':     (';-)', ';)', ';D'),
                  'poke.gif':     (':-P', ':P', '=P', '=p', '=b', ':-p', ':p', ':-b', ':b')}

# these are used for passing record fields
class _record: pass

manage_addZUBBFolderForm = HTMLFile('dtml/ZUBBFolder_addForm', globals())
manage_addZUBBMessageForm = HTMLFile('dtml/ZUBBMessage_addForm', globals())

def manage_addZUBBFolder(self, id, title, REQUEST=None):
    """Add a ZUBBFolder object"""

    folder = ZUBBFolder(id, title)
    self._setObject(id, folder)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

class ZUBBMessage(OFS.SimpleItem.SimpleItem,
             OFS.PropertyManager.PropertyManager,
             Persistent,
             Implicit,
             AccessControl.Role.RoleManager):
    """ZUBBMessage object"""

    meta_type = "ZUBB Message"

    icon = 'misc_/ZUBB/zubb_message.gif'

    _properties = (
        {'id':'title', 'type':'string', 'mode':'w'},
        {'id':'author', 'type':'string', 'mode':'w'},
        {'id':'email', 'type':'string', 'mode':'w'},
        {'id':'body', 'type':'text', 'mode':'w'},
        {'id':'in_reply_to', 'type':'string', 'mode':''},
        {'id':'priority', 'type':'integer', 'mode':'w'}
        )

    manage_options=(
        {'label': 'Contents',
         'action': 'manage_main'},
        
        {'label': 'View',
         'action': ''},
        
        {'label': 'Security',
         'action': 'manage_access'}
        )


    security = ClassSecurityInfo()

    security.declareProtected('Manage ZUBB Folders', 'manage_addProperty',
                                                    'manage_editProperties',
                                                    'manage_delProperties',
                                                    'manage_changeProperties')

    def __init__(self, id, title, body, author, email, in_reply_to, priority = 0):
        self.id = id
        self.title = title.strip()
        self.author = author.strip()
        self.email = email.strip()
        self.body = body
        self.in_reply_to = in_reply_to
        self.priority = int(priority)
        self.posted_date = self.last_post_date = DateTime()
        self.last_post_author = ''

    security.declareProtected('Manage ZUBB Folders', 'manage_main')
    manage_main = HTMLFile('dtml/ZUBBMessage_editForm', globals()) # Management Interface

    security.declarePublic('index_html')
    def index_html(self):
        """ standard_view """
        if hasattr(self.aq_parent, 'zubb_options'):
            zubb_options = self.aq_parent.zubb_options()
        else:
            zubb_options = {}
        # default view to the primary_view
        view = zubb_options.get('primary_view', 1)
        req = self.REQUEST
        if req:
            # see if 'view' is a form variable
            if req.form.has_key('view'):
                view = int(req.form['view'])
                # set a cookie
                req.RESPONSE.setCookie('zubbView', view, path='/')
            # otherwise try getting 'view' from cookies
            elif req.cookies.has_key('zubbView'):
                view = int(req.cookies['zubbView'])
        if view == 2:
            return self.message_inline(None, self)
        else:
            return self.message_indent(None, self)

    security.declarePublic('thread_top')
    def thread_top(self):
        """ returns the top ancestor """
        top = self
        irp = top.in_reply_to
        folder = self.aq_parent.aq_explicit
        while irp:
            top = getattr(folder, irp)
            irp = top.in_reply_to
        return top

    security.declarePublic('ancestors')
    def ancestors(self):
        """Return a list of ancestral posts"""
        ancestors = []
        irp = self.in_reply_to
        folder = self.aq_parent.aq_explicit
        while irp:
            parent = getattr(folder, irp)
            ancestors.append(parent)
            irp = parent.in_reply_to
        return ancestors

    security.declarePublic('children')
    def children(self):
        """ Return a list of replies to this message - most recently posted first """
        children = []
        results = self.searchResults(in_reply_to=self.id, sort_on='last_post_date', sort_order='reverse')
        children = map(lambda x: x.getObject(), results)
        return children

    security.declarePublic('descendants')
    def descendants(self):
        """ Return a list of descendants and the date they were posted (including self) """
        # add self
        descendants = [(self.posted_date, self)]
        # add children
        for child in self.children():
            descendants = descendants + child.descendants()
        return descendants

    security.declarePublic('inline')
    def inline(self):
        """ Return all the descendant posts inline (flatfile style) """
        descendants = self.descendants()
        descendants.sort()   # now it's in order by posted date
        self_irp = self.in_reply_to
        thread = []
        threadidx = []
        index = -1
        for descendantdate, descendant in descendants:
            index = index + 1
            irp = descendant.in_reply_to
            if irp == self_irp:
                reply = -1
            else:
                reply = threadidx.index(irp)
            thread.append((descendant, index, reply))
            threadidx.append(descendant.getId())
        return thread

    security.declarePublic('indent')
    def indent(self):
        """ Return all the descendant posts tree style """
        self_irp = self.in_reply_to
        thread = []
        threadidx = []
        index = -1
        for descendantdate, descendant in self.descendants():
            index = index + 1
            irp = descendant.in_reply_to
            if irp == self_irp:
                reply = -1
                indent = 0
            else:
                reply = threadidx.index(irp)
                indent = thread[reply][1] + 1
            thread.append((descendant, indent))
            threadidx.append(descendant.getId())
        return thread

    security.declareProtected('Manage ZUBB Folders', 'manage_edit')
    def manage_edit(self, REQUEST):
        """Post a ZUBB Message"""
        f = REQUEST.form
        title = f.get('title', '')
        body = f.get('body', '')
        in_reply_to = f.get('in_reply_to', '')
        priority = f.get('priority', '')
        ok = 0
        if in_reply_to:
            if not hasattr(self.aq_parent.aq_explicit, in_reply_to):
                msg = 'Not updated: "In Reply To" ID did not exist.'
            elif in_reply_to == self.id:
                msg = 'Not updated: tried to reference self for "In Reply To".'
            else:
                ok = 1
        else:
            ok = 1
        if ok and (title or body):
            self.title = title
            self.body = body
            self.author = f.get('author', '')
            self.email = f.get('email', '')
            self.in_reply_to = in_reply_to
            if priority:
                self.priority = 1
            else:
                self.priority = 0
            self._catalog()
            msg = "ZUBB message updated."
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, manage_tabs_message = msg)

    def manage_afterAdd(self, item, container):
        self.no_orphan_fixing = 0   # allow thread flattening
        OFS.SimpleItem.SimpleItem.manage_afterAdd(self, item, container)
        self._update_thread(self)   # <-- catalog done in here

    def manage_beforeDelete(self, item, container):
        self.no_orphan_fixing = 1   # so that we wont flatten our thread if this is a move / rename
        parent = self.aq_parent
        id = self.getId()
        self.uncatalog_object(id)
        if not getattr(parent,'no_orphan_fixing',0):
            # Prevent orphans
            for child in self.children():
                child.in_reply_to = self.in_reply_to
                child._catalog()
            # Set new last post for parent if we need to
            parent = getattr(self.aq_parent, self.in_reply_to)
            if parent.last_post_date == self.posted_date:
                parent._update_thread(parent.thread_last_post(id))   # find newest message (ignoring current one)
        OFS.SimpleItem.SimpleItem.manage_beforeDelete(self, item, container)

    security.declarePrivate('_catalog')
    def _catalog(self):
        """ Catalog this message """
        self.catalog_object(self, self.getId())

    security.declarePublic('childMessageCount')
    def childMessageCount(self):
        """ Counts the number of children of this message, including this message"""
        count = 1   # this message
        for message in self.children():
                count = count + message.childMessageCount()
        return count

    security.declarePublic('threadMessageCount')
    def threadMessageCount(self):
        """ Looks at the topmost ancestor, and then counts children"""
        message_count = self.thread_top().childMessageCount()
        return message_count

    security.declarePublic('child_last_post')
    def child_last_post(self, ignore_deleted=''):
        """ Looks through children and returns the newest message"""
        last_post_date = self.posted_date
        last_post = self
        for message in self.children():
            if message.id != ignore_deleted:
                test_post = message
                test_post_date = message.child_last_post()
                if test_date > last_post_date:
                    last_post = test_post
                    last_post_date = test_post_date
        return last_post

    security.declarePublic('thread_last_post')
    def thread_last_post(self, ignore_deleted=''):
        """ Looks at the topmost ancestor, and then recurses
            through the messages finding the last posted message"""
        last_post = self.thread_top().child_last_post(ignore_deleted)
        return last_post

    security.declarePublic('all_text')
    def all_text(self):
        """ Returns the title and the body text - used to index for search """
        return '%s %s' % (self.title, self.body)

    security.declarePrivate('_update_thread')
    def _update_thread(self, last_post):
        """ Changes the last post down the family tree """
        last_post_date = last_post.posted_date
        last_post_author = last_post.author
        self.last_post_date = last_post_date
        self.last_post_author = last_post_author
        self._catalog()
        irp = self.in_reply_to
        if irp:
            getattr(self.aq_parent, irp)._update_thread(last_post)


catalog_indexes =   ( ('author', 'FieldIndex')
                    , ('all_text', 'ZCTextIndex')
                    , ('email', 'FieldIndex')
                    , ('in_reply_to', 'FieldIndex')
                    , ('last_post_date', 'FieldIndex')
                    , ('last_post_author', 'FieldIndex')
                    , ('posted_date', 'FieldIndex')
                    , ('priority', 'FieldIndex')
                    )

catalog_metadata =   ('author'
                    , 'email'
                    , 'id'
                    , 'in_reply_to'
                    , 'last_post_date'
                    , 'last_post_author'
                    , 'posted_date'
                    , 'priority'
                    , 'title'
                    )

class ZUBBFolder(ZCatalog.ZCatalog):
    """ZUBBFolder object"""

    meta_type = 'ZUBB Folder'

    security = ClassSecurityInfo()

    security.setPermissionDefault('Add ZUBB Messages', ('Anonymous','Manager'))
    security.setPermissionDefault('Delete ZUBB Folders', ('Manager',))
    security.setPermissionDefault('Manage ZUBB Folders', ('Manager',))

    security.declareProtected('View management screens', 'manage_catalogOptions')
    manage_catalogOptions = Globals.HTMLFile('dtml/catalogOptions', globals())

    security.declareProtected('Manage ZUBB Folders', 'manage_ZUBBOptions')
    manage_ZUBBOptions = Globals.HTMLFile('dtml/ZUBBOptions', globals())

    manage_options=(
        {'label': 'Contents',
         'action': 'manage_main'},
        
        {'label': 'View',
         'action': ''},
        
        {'label': 'ZUBB Options',
         'action': 'manage_ZUBBOptions'},
        
        {'label': 'Catalog Options',
         'action': 'manage_catalogOptions'},
        
        {'label': 'Security',
         'action': 'manage_access'},

        {'label': 'Undo',
         'action': 'manage_UndoForm'}
        )

    def __init__(self, id, title=''):
        ZCatalog.ZCatalog.__init__(self, id, title)   # Initialise ZCatalog
        self._indexes = 0
        self._defaultfiles = 0
        base = aq_base(self)
        for column_name in catalog_metadata:
            self.addColumn(column_name)
        # indexes are set up in manage_afterAdd

    security.declarePublic('zubb_options')
    def zubb_options(self):
        """ Return ZUBB Folder options (inherit if necessary) """
        options = {}
        inherit = getattr(self, '_option_inherit', 1)
        if inherit:
            if hasattr(self.aq_parent, 'zubb_options'):
                ancestor_options = self.aq_parent.zubb_options()
                if type(ancestor_options) == type(options):
                    options = ancestor_options
        options['inherit'] = inherit
        if not options.has_key('emoticons_directory'):
            options['emoticons_directory'] = getattr(self, '_option_emoticons_directory', '/misc_/ZUBB')
        if not options.has_key('date_structure'):
            options['date_structure'] = getattr(self, '_option_date_structure', '%d&nbsp;%b&nbsp;%Y')
        if not options.has_key('date_friendly'):
            options['date_friendly'] = getattr(self, '_option_date_friendly', 1)
        if not options.has_key('primary_view'):
            options['primary_view'] = getattr(self, '_option_primary_view', 1)
        if not options.has_key('secondary_view'):
            options['secondary_view'] = getattr(self, '_option_secondary_view', 2)
        if not options.has_key('skin'):
            if not hasattr(self, 'skins'):
                zubb_files.addFolder(self, 'skins')
            options['skin'] = getattr(self, '_option_skin', self['skins'])
        if not options.has_key('search_method'):
            options['search_method'] = getattr(self, '_option_search_method', 0)
        return options

    security.declareProtected('Manage ZUBB Folders', 'manage_saveOptions')
    def manage_saveOptions(self, date_structure, primary_view, secondary_view, emoticons_directory, date_structure_custom = '', emoticons_directory_custom = '', date_friendly = 1, inherit = 1, skin_id = '', search_method = 0, REQUEST = None):
        """ Save ZUBB Folder options """
        if date_structure_custom:
            date_structure = date_structure_custom
        if date_structure:
            self._option_date_structure = date_structure
        if emoticons_directory_custom:
            emoticons_directory = emoticons_directory_custom
        if emoticons_directory:
            self._option_emoticons_directory = emoticons_directory
        self._option_primary_view = primary_view
        if secondary_view == primary_view:
            secondary_view = 0
        self._option_secondary_view = secondary_view
        self._option_date_friendly = date_friendly
        self._option_inherit = inherit
        try:
            self._option_skin = self['skins'][skin_id]
        except:
            if not hasattr(self, 'skins'):
                zubb_files.addFolder(self, 'skins')
            self._option_skin = self['skins']
        self._option_search_method = search_method
        self._p_changed = 1   # updates the folder object properly
        if REQUEST:
            return self.manage_main(self, REQUEST, manage_tabs_message = "ZUBB Folder options saved.")

    security.declarePublic('zubb_emoticons')
    def zubb_emoticons(self):
        return emoticons_list
    
    security.declarePublic('emoticon_html')
    def emoticon_html(self, file, emoticon, directory = ''):
        """ Return html for the emoticon image """
        emoticon = html_quote(emoticon)
        if not directory:
            directory = getattr(self.zubb_options(), 'emoticons_directory', '/misc_/ZUBB')
        return '<img src="%s/%s" alt="%s" />' % (directory, file, emoticon)

    security.declareProtected('Manage ZUBB Folders', 'manage_editZUBBFolder')
    def manage_editZUBBFolder(self, title, REQUEST=None):
        """Set properties of ZUBBFolder.  Presently just title."""
        self.title = title
        if REQUEST:
            return self.manage_ZUBBOptions(self, REQUEST, manage_tabs_message = "ZUBB Folder title changed.")

    security.declarePublic('posts', 'children')
    def posts(self):
        """ Return the top-level posts - most recently posted into first """
        posts = []
        results = self.searchResults(in_reply_to=[''], sort_on='last_post_date', sort_order='reverse', priority=1)
        results += self.searchResults(in_reply_to=[''], sort_on='last_post_date', sort_order='reverse', priority=0)
        posts = map(lambda x: x.getObject(), results)
        return posts

    children = posts

    security.declarePublic('forum_url')
    def forum_url(self):
        return '/%s' % self.absolute_url(1)
    
    security.declarePublic('generate_id')
    def generate_id(self):
        """Find an available ID"""
        fmt = "%08d"
        id = 0
        modifier = 1
        direction = -1
        increment = 1
        hasit = hasattr(self.aq_explicit, fmt % id)
        while modifier != 1 or hasit:
            # check direction
            if hasit:
                direction = 1
            else:
                if increment:
                    increment = 0
                direction = -1
            # change modifier
            if increment:
                modifier = modifier * 2
            else:
                if modifier > 1:
                    modifier = modifier / 2
            # change id by modifier
            id = id + (direction * modifier)
            hasit = hasattr(self.aq_explicit, fmt % id)
        return fmt % id

    security.declareProtected('Add ZUBB Messages', 'do_post')
    def do_post(self, REQUEST):
        """Post a ZUBB Message"""
        f = REQUEST.form
        title = f.get('title', '')
        body = f.get('body', '')
        author = f.get('author', '')
        email = f.get('email', '')
        in_reply_to = f.get('in_reply_to', '')
        priority = f.get('priority', 0)
        # set cookies
        resp = REQUEST.RESPONSE
        if author:
            resp.setCookie('zubbAuthor', author, path='/')
        else:
            resp.expireCookie('zubbAuthor')
        if email:
            resp.setCookie('zubbEmail', email, path='/')
        else:
            resp.expireCookie('zubbEmail')
        self.manage_addZUBBMessage('', title, body, author, email, in_reply_to, priority, REQUEST)

    security.declareProtected('Add ZUBB Messages', 'manage_addZUBBMessage')          
    def manage_addZUBBMessage(self, id, title, body, author, email, in_reply_to='', priority=0, REQUEST=None):
        """Add a ZUBBMessage to this ZUBBFolder"""
        title = title.strip()
        body.strip()
        # only add message if there is a title or some body text
        if (title!='' and title!='RE:') or body!='':
            if not id:
                id = self.generate_id()
            message = ZUBBMessage(id, title, body, author, email, in_reply_to, priority)
            self._setObject(id, message)
        else:
            id = None
        if REQUEST is not None:
            if not id:
                # didn't add message
                REQUEST.RESPONSE.redirect(REQUEST.URL1)
            else:
                # redirect
                if REQUEST.form.has_key('goto'):
                    REQUEST.RESPONSE.redirect('%s' % REQUEST.form.goto)
                else:
                    REQUEST.RESPONSE.redirect('%s' % id)

    security.declareProtected('Delete ZUBB Messages', 'do_delete')
    def do_delete(self, REQUEST):
        """Delete a ZUBB Message"""
        id = REQUEST.form.get('delete')
        message = ''
        if id:
            if hasattr(self, id):
                try:
                    self.manage_deleteZUBBMessage(id)
                    message = 'Post deleted'
                except:
                    message = 'Could not delete: an error occured while trying to delete post'
            else:
                message = 'Could not delete: post not found'
        querystring = self.querystring_change(REQUEST.QUERY_STRING, ('delete',('message', message)))
        if REQUEST:
            REQUEST.RESPONSE.redirect('%s%s' % (REQUEST.URL1, querystring))
            
    security.declareProtected('Delete ZUBB Messages', 'manage_deleteZUBBMessage')
    def manage_deleteZUBBMessage(self, id, REQUEST=None):
        """Delete a ZUBBMessage from this ZUBBFolder"""
        if hasattr(self, id):
            if isinstance(self[id], ZUBBMessage):
                self.manage_delObjects([id])
        if REQUEST:
            REQUEST.RESPONSE.redirect('%s/manage_main' % REQUEST.URL1)

    def manage_beforeDelete(self, item, container):
        self.no_orphan_fixing = 1   # so that we wont flatten our thread if this is a move / rename
        ZCatalog.ZCatalog.manage_beforeDelete(self, item, container)

    def manage_afterAdd(self, item, container):
        self.no_orphan_fixing = 0   # allow thread flattening
        ZCatalog.ZCatalog.manage_afterAdd(self, item, container)
        # set up the indexes
        if not self._indexes:
            casenormalizer = _record()
            stopwords = _record()
            recordsplitter = _record()
            casenormalizer.group = 'Case Normalizer'
            casenormalizer.name = 'Case Normalizer'
            stopwords.group = 'Stop Words'
            stopwords.name = ' Don\'t remove stop words'
            recordsplitter.group = 'Word Splitter'
            recordsplitter.name = 'Whitespace splitter'
            records = [casenormalizer, stopwords, recordsplitter]
            self.manage_addProduct['ZCTextIndex'].manage_addLexicon('lexicon', 'Stores message text for searching', elements = records)
            extra = _record()
            extra.index_type = 'Okapi BM25 Rank'
            extra.lexicon_id = 'lexicon'
            for index_name, index_type in catalog_indexes:
                if index_type == 'ZCTextIndex':
                    extra.doc_attr = index_name
                    self.addIndex(index_name, index_type, extra)
                else:
                    self.addIndex(index_name, index_type)
            self._indexes = 1
        # add the default files unless our parent is also a ZUBB folder
        if not self._defaultfiles:
            if self.aq_parent.meta_type != 'ZUBB Folder':
                self.manage_addDefaultFiles()
            self._defaultfiles = 1

    security.declareProtected('Manage ZUBB Folders', 'manage_recatalog')
    def manage_recatalog(self, REQUEST = None):
        """ Recatalog all messages in the catalog """
        self.manage_catalogClear()
        results = self.objectItems(['ZUBB Message'])
        for id, ob in results:
            ob._catalog()
        if REQUEST:
            return self.manage_main(self, REQUEST, manage_tabs_message = "ZUBB Folder Recatalogued.")

    security.declareProtected('Manage ZUBB Folders', 'manage_addDefaultFiles')
    manage_addDefaultFiles = ZUBB_files.manage_addDefaultFiles

    security.declareProtected('Manage ZUBB Folders', 'manage_regenerateDefaultMethods')
    manage_regenerateDefaultMethods = HTMLFile('dtml/regenerateDefaultMethods', globals())

    security.declarePublic('get_batch'
                         , 'text_limit'
                         , 'text_format'
                         , 'querystring_change'
                         , 'date_format'
                         , 'get_subfolders'
                    )
    get_batch = ZUBB_methods.get_batch
    text_limit = ZUBB_methods.text_limit
    text_format = ZUBB_methods.text_format
    querystring_change = ZUBB_methods.querystring_change
    date_format = ZUBB_methods.date_format
    get_subfolders = ZUBB_methods.get_subfolders

    security.declarePublic('search_messages')
    def search_messages(self, words, author = '', threads_only = None, weighted = None):
        """ Returns all of the messages that contain the words specified """
        # search for partial words
        if words:
            words_list = words.split(' ')
            words_list = map(lambda x: x and '%s*' % x, words_list)
            words = ' '.join(words_list)
        # search for title and author
        if weighted == None:
            directory = getattr(self.zubb_options(), 'search_method', 0)
        if weighted == 1:
            # return results by weight
            results = self.searchResults(all_text = words, author = author)
        else:
            # return results by most recent
            results = self.searchResults(all_text = words, author = author, sort_on='posted_date', sort_order='reverse')
        if threads_only:
            # take only the tops of threads where the word occured and avoid double ups
            newresults = []
            for result in results:
                if result.in_reply_to:
                    newresult = result.thread_top
                else:
                    newresult = result
                if newresult not in newresults:
                    newresults.append(newresult)
            results = newresults
        return results

# This gets the security assertions going
Globals.InitializeClass(ZUBBMessage)
Globals.InitializeClass(ZUBBFolder)
