import re

def get_batch(self, results, size=25, orphan=2):
    """ Gets one batch of results """
    from ZTUtils import Batch
    REQUEST = self.REQUEST

    # get the size of all the results
    last = len(results) - 1

    # set the batch variables
    REQUEST.set('batch_record_count',last + 1)

    if last > orphan:
      batch_end = int((last - orphan)/size)
    else:
      batch_end = 0
    REQUEST.set('batch_end',batch_end)

    if not hasattr(REQUEST,'batch_current'):
      REQUEST.set('batch_current','0')
      batch_current = 0
    else:
      if REQUEST.batch_current=='last':
        REQUEST.set('batch_current',batch_end)
      batch_current = int(REQUEST.batch_current)

    if batch_current > 0:
      REQUEST.set('batch_prev',batch_current-1)

    if batch_current < batch_end:
      REQUEST.set('batch_next',batch_current+1)

    # set the start / end of the batch we actually want
    start = size * batch_current
    
    if start + size - 1 + orphan < last:
      end = start + size - 1
    else:
      end = last

    # get the batch we want
    batch = results[start:end + 1]

    return batch

def text_limit(self, text, maxlength = 30):
    """ Limit text to a maximum length """
    if len(text) > maxlength:
        return '%s...' % text[:maxlength - 2]
    else:
        return text

def querystring_change(self, querystring, items):
    """ Add / remove items from the querystring
        To add, make the item a list or tuple like this: ('name', 'value')
        To remove, just make the item the string you want to remove: 'name' """
    from Products.PythonScripts.standard import url_quote

    if querystring and querystring[0] == '?':
      querystring = querystring[1:]

    if querystring == '':
      querylist = []
    else:
      querylist = querystring.split('&')

    for item in items:
      itemvalue = None
      if (type(item) == type(())) or (type(item) == type([])):
        itemname = item[0]
        if len(item) > 1:
          itemvalue = item[1]
      else:
        itemname = item
      if itemvalue:
        #add item to querystring
        itemvalue = url_quote(str(itemvalue))
        added = 0
        for index in range(len(querylist)):
          if querylist[index].split('=')[0] == itemname:
            querylist[index] = '%s=%s' % (itemname,itemvalue)
            added = 1
            break
        if not added:
          querylist.append('%s=%s' % (itemname,itemvalue))
      else:
        #remove item from querystring
        for index in range(len(querylist)):
          if querylist[index].split('=')[0] == itemname:
            del querylist[index]
            break

    querystring = '&'.join(querylist)
    if querystring:
      querystring = '?%s' % querystring

    return querystring


url_regexp_headers = "(https?|ftps?)://"   # http:// https:// ftp:// ftps://
url_regexp_domain = r"((([0-9]{1,3}\.){2})?([0-9]{1,3}\.){3}[0-9]{1,3}"   # IP 199.194.52.184 (or IPV6)
url_regexp_domain += "|"   # either IP or domain name
url_regexp_domain += r"([0-9a-z~][-0-9a-z_!~*'()]*\.)*"   # tertiary domain(s)- www.
url_regexp_domain += r"([0-9a-z][-0-9a-z]{0,61})?[0-9a-z]\."   # second level domain 
url_regexp_domain += r"[a-z]{2,6})"   # first level domain- .com or .museum 
url_regexp_tail = r"(((/[-0-9a-z_!~*'().;?:@&=+$,%#]+)+/?)"   # filename/queries/anchors
url_regexp_tail += r"|(/?))"   # a slash isn't required if there is no file name 

url_regexp_strict_tail = r"(((/[-0-9a-z_!~*'().;?:@&=+$,%#]*[-0-9a-z_~*'@&=+$%#/])+)"   # filename/queries/anchors - don't end in punctuation
#url_regexp_strict_tail = r"((/[0-9a-z_~*'@&=+$%#\-/]+)+)"   # filename/queries/anchors - don't end in punctuation
url_regexp_strict_tail += r"|(/?))"   # a slash isn't required if there is no file name 

url_regexp = "(" + url_regexp_headers + ")?"
url_regexp += r"(([-0-9a-z_!~*'().&=+$%]+:)?[-0-9a-z_!~*'().&=+$%]+@)?"   # username:password@ 
url_regexp += url_regexp_domain
url_regexp += r"(:[0-9]{1,4})?"   # port number- :80
url_regexp += url_regexp_tail

url_regexp_strict = "((" + url_regexp_headers + ""
url_regexp_strict += r"((([-0-9a-z_!~*'().&=+$%]+:)?[0-9a-z_!~*'().&=+$%-]+@)?))|"   # username:password@ (only if using a header)
url_regexp_strict += r"[0-9a-z~][-0-9a-z_!~*'()]*\.)"   # we need a tertiary if a header wasn't provided
url_regexp_strict += url_regexp_domain
url_regexp_strict += r"(:[0-9]{1,4})?"   # port number- :80
url_regexp_strict += url_regexp_strict_tail   # don't end in punctuation

def text_format(self, text):
    """ Format the posts """
    from Products.PythonScripts.standard import html_quote, newline_to_br
    from ZUBB import emoticons_list

    # html quote the text
    text = html_quote(text)

    # replace emoticons with images
    for file, emoticons in emoticons_list.items():
        for emoticon in emoticons:
            image_html = self.emoticon_html(file, emoticon)
            regexp_emoticon = re.escape(emoticon)
            regexp = re.compile(r'(((?<=[\s<>])' + regexp_emoticon + ')|(' + regexp_emoticon + r'(?=($|[\s<>]))))')
            if regexp.search(text):
                text = regexp.sub(image_html, text)

    # [url]www.example.com[/url]  or  [url=www.example.com]name[/url]
    def url_block_replace(matchobj):
        block = matchobj.group(0)
        block_header = re.search(regexp_block_header('url'), block).group(0)
        regexp = re.compile(r'(?<=\]).*(?=\[)', re.DOTALL)
        block_contents = regexp.search(block).group(0).strip()
        url = re.search(r'(?<=[=\s])' + url_regexp + r'(?=[\s\]])', block_header)
        if not url:
            url = re.search(r'(?<=(=|\s))' + url_regexp_tail + r'(?=(\s|\]))', block_header)
            if not url:
                url = re.search(r'(?<=(\s|\]))' + url_regexp + r'(?=(\s|\[))', block_contents)
                if not url:
                    url = re.search(r'(?<=(\s|\]))' + url_regexp_tail + r'(?=(\s|\[))', block_contents)
        if url:
            url = url.group(0)
            url = add_url_header(url)
            block = re.sub('^' + regexp_block_header('url') + r'\s*', '<a href="%s">' % url, block)
            block = re.sub(r'\s*' + regexp_block_footer('url') + '$', '</a>', block)
        return block
    text = replace_block(text, 'url', url_block_replace, no_nested = 1)

    # process emails
    def email_replace(matchobj):
        email = matchobj.group(0)
        email = '<a href="mailto:%s">%s</a>' % (email, email)
        return email
    #regexp = r"[0-9a-z_!~*'-]+(\.[0-9a-z_!~*'-]+)*@" + url_regexp_domain
    regexp = r"(?<![^\s(\[\]{;,])[0-9a-z_!~*'-]+(\.[0-9a-z_!~*'-]+)*@" + url_regexp_domain + r'(?=([\s)\[\]}:;?!.,&]|$))'   # must start with white space and end with white space or end of string
    text = re.sub(regexp, email_replace, text)

    # process urls
    def url_replace(matchobj):
        url = matchobj.group(0)
        url = '<a href="%s">%s</a>' % (add_url_header(url), url)
        return url
    #regexp = r'(?<!")' + url_regexp
    regexp = r'(?<![^\s(\[\]{:;,])' + url_regexp_strict + r'(?=([\s)\[\]}:;?!.,]|$))'   # must start with white space and end with white space or end of string
    text = re.sub(regexp, url_replace, text)

    # [quote]
    def quote_replace(matchobj):
        block = matchobj.group(0)
        block = re.sub(regexp_block_header('quote') + r'\s*', '<div class="quotehead">Quote:</div><div class="quoteblock">', block, 1)
        block = re.sub(r'\s*' + regexp_block_footer('quote'), '</div>', block, 1)
        return block
    text = replace_block(text, 'quote', quote_replace)

    # [code]
    def quote_replace(matchobj):
        block = matchobj.group(0)
        # keep it space indented
        block = re.sub('(?<=\n) +', lambda x: x.group(0).replace(' ', '&nbsp;'), block)
        # replace blocks with div
        block = re.sub(r'%s\s*' % regexp_block_header('code'), '<div class="codeblock">', block, 1)
        block = re.sub(r'\s*%s' % regexp_block_footer('code'), '</div>', block, 1)
        return block
    text = replace_block(text, 'code', quote_replace)

    # [b]
    def b_replace(matchobj):
        block = matchobj.group(0)
        block = re.sub(regexp_block_header('b') + r'\s*', '<strong>', block, 1)
        block = re.sub(r'\s*' + regexp_block_footer('b'), '</strong>', block, 1)
        return block
    text = replace_block(text, 'b', b_replace)

    # [i]
    def i_replace(matchobj):
        block = matchobj.group(0)
        block = re.sub(regexp_block_header('i') + r'\s*', '<em>', block, 1)
        block = re.sub(r'\s*' + regexp_block_footer('i') , '</em>', block, 1)
        return block
    text = replace_block(text, 'i', i_replace)
    
    # [image]www.example.com/example.gif[/image]  or  [image=www.example.com/example.gif]alt tag[/image]
    def image_block_replace(matchobj):
        block = matchobj.group(0)
        block_header = re.search(regexp_block_header('image'), block).group(0)
        regexp = re.compile(r'(?<=\]).*(?=\[)', re.DOTALL)
        block_contents = regexp.search(block).group(0).strip()
        url = re.search(r'(?<=(=|\s))' + url_regexp + r'(?=(\s|\]))', block_header)
        if not url:
            url = re.search(r'(?<=(=|\s))' + url_regexp_tail + r'(?=(\s|\]))', block_header)
        alt = ''
        if not url:
            url = re.search(r'(?<=(\s|\]))' + url_regexp + r'(?=(\s|\[))', block_contents)
            if not url:
                url = re.search(r'(?<=(\s|\]))' + url_regexp_tail + r'(?=(\s|\[))', block_contents)
        else:
            #strip out any html tags in the alt tag
            alt = re.sub('<.*?>.*?<.*?>', '', block_contents)
        if url:
            url = url.group(0)
            url = add_url_header(url)
            block = '<img src="%s" alt="%s" />' % (url, alt)
        return block
    text = replace_block(text, 'image', image_block_replace, no_nested=1, allow_empty=1)
    
    # put in the <br>s
    text = newline_to_br(text)

    return text

def regexp_block(block, allow_empty = 0, no_nested = 0):
    if no_nested:
        if allow_empty:
            content = r'[^\[\]]*?'
        else:
            content = r'[^\[\]]*?[^\[\]\s]+?[^\[\]]*?'
    else:
        if allow_empty:
            content = '.*'
        else:
            content = r'.*\S+.*'
    return regexp_block_header(block) + content + regexp_block_footer(block)

def regexp_block_header(block):
    return r'\[\s*' + block +  r'([=\s].*?)?\]'

def regexp_block_footer(block):
    block = re.escape(block)
    return r'\[\s*/\s*' + block + r'\s*\]'

def add_url_header(url):
    """ Add the default http:// if the url doesn't have a header (and doesn't start with a /) """
    if not re.search('^(' + url_regexp_headers + '|/)', url):
        return 'http://%s' % url
    else:
        return url

def replace_block(text, tag, method, allow_empty = 0, no_nested = 0):
    regexp = re.compile(regexp_block(tag, allow_empty = allow_empty, no_nested = no_nested), re.DOTALL)
    newtext = regexp.sub(method, text)
    if newtext != text:
        text = replace_block(newtext, tag, method)
    return text


def date_format(self, date):
    """ Format the date with the correct format """
    from DateTime import DateTime
    zubb_options = self.zubb_options()
    if zubb_options['date_friendly']:
        now = DateTime().strftime('%d%m%Y')
        if date.strftime('%d%m%Y') == now:
            return date.strftime('Today&nbsp;%I:%M&nbsp;%p')
        elif (date+1).strftime('%d%m%Y') == now:
            return date.strftime('Yesterday&nbsp;%I:%M&nbsp;%p')
    date_structure = zubb_options['date_structure']
    date_structure = date_structure.replace(' ', '&nbsp;')
    return date.strftime(date_structure + '&nbsp;%I:%M&nbsp;%p')

def get_subfolders(self, meta_type = 'ZUBB Folder'):
    folders = list(self.objectValues(meta_type))
    folders = map(lambda x: (x.title_or_id(), x), folders)
    folders.sort()
    folders = map(lambda x: x[1], folders)
    return folders
